/* netjail.c - The entire source for the netjail library */

/* Copyright (C) 2004 Craig B. Agricola

   This file is part of the netjail library for limiting net access in
   user space.

   netjail is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.

   netjail is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.

   You should have received a copy of the GNU General Public License along
   with Bash; see the file COPYING.  If not, write to the Free Software
   Foundation, 59 Temple Place, Suite 330, Boston, MA 02111 USA. */

#include <sys/socket.h>
#include <dlfcn.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <errno.h>
#include <string.h>
#include <stdarg.h>
#include <netdb.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "config.h"

static int lib_initialized = 0;
static int (*orig_socket)(int, int, int) = 0;
static int (*orig_connect)(int, const struct sockaddr*, socklen_t) = 0;
static FILE *log = 0;
static int debug_rules = 0;

typedef enum {ALLOW=1, DENY=2} AD_policy_t;

struct INET_allow_deny {
  AD_policy_t              AD_policy;
  struct in_addr           addr;
  struct in_addr           mask;
  in_port_t                port_range_start;
  in_port_t                port_range_end;
  struct INET_allow_deny  *next;
};
static struct INET_allow_deny *INET_AD_list = 0;
static struct INET_allow_deny *INET_AD_tail = 0;

struct UNIX_allow_deny {
  AD_policy_t              AD_policy;
  int                      all;
  char                    *path;
  struct UNIX_allow_deny  *next;
};
static struct UNIX_allow_deny *UNIX_AD_list = 0;
static struct UNIX_allow_deny *UNIX_AD_tail = 0;

static unsigned int INETfds[32];
static unsigned int UNIXfds[32];
#define addINETfd(fd) { INETfds[fd>>5] |= (1<<(fd&0x1f)); }
#define addUNIXfd(fd) { UNIXfds[fd>>5] |= (1<<(fd&0x1f)); }
#define isINETfd(fd)  ((INETfds[(fd)>>5]>>((fd)&0x1f))&0x1)
#define isUNIXfd(fd)  ((UNIXfds[(fd)>>5]>>((fd)&0x1f))&0x1)

void die(char *fmt, ...) {
  va_list args;
  va_start(args, fmt);
  vfprintf(stderr, fmt, args);
  va_end(args);
  fprintf(stderr, "\n");
  fflush(stderr);
  exit(-1);
}

void log_msg(char *fmt, ...) {
  va_list args;
  time_t t;
  char *tstr, *c;

  time(&t);
  tstr = strdup(ctime(&t));
  for(c=tstr; (*c!='\n')&&(*c!=0); c++);
  if(*c=='\n'){*c=0;}
  fprintf(log, "%s: ", tstr);
  free(tstr);

  va_start(args, fmt);
  vfprintf(log, fmt, args);
  va_end(args);

  fflush(log);
}

static char *parse_cb_error = 0;
static int   parse_cb_error_char = 0;
#define throw_parse_cb_error(msg,where)                           \
{                                                                 \
  parse_cb_error = msg;                                           \
  parse_cb_error_char = ((where)==-1)?(walk-addr_spec+1):(where); \
  return(1);                                                      \
}

int parse_dotted_quad_address(char *addr_spec, struct in_addr *addr) {
  int dotcount = 0;
  char *walk, *quadstart = addr_spec;
  int quadval;
  addr->s_addr = 0;
  for(walk=addr_spec; dotcount<4; walk++) {
    if ((*walk=='.') || (*walk==0)) {
      quadval = atoi(quadstart);
      if ((quadval<0) || (quadval>255))
        throw_parse_cb_error("Dotted-quad value out of range [0..255]",
                             quadstart - addr_spec);
      addr->s_addr |= ((quadval & 0xff) << (dotcount*8));
      quadstart = walk + 1;
      dotcount++;
      if (*walk==0) { dotcount=4; }
      *walk=0;
    }
  }
  return(0);
}

int parse_INET_AD_cb(AD_policy_t AD_policy, char *addr_spec) {
  char *walk, *mask_str=0, *ports_str=0;
  int addr_kind;
  struct in_addr addr;
  struct in_addr mask;
  in_port_t      port_range_start;
  in_port_t      port_range_end;

  struct INET_allow_deny *new;

  /* Scan for specification parts */
  for(walk=addr_spec; *walk!=0; walk++) {
    if        (*walk=='/') {
      if (mask_str) throw_parse_cb_error("Found a second net mask", -1);
      if (ports_str)
        throw_parse_cb_error("Found a net mask after the ports", -1);
      *walk=0; mask_str = walk + 1;
    } else if (*walk==':') {
      if (ports_str)
        throw_parse_cb_error("Found a second port specification", -1);
      *walk=0; ports_str = walk + 1;
    }
  }

  /* Try to digest the address part */
  addr_kind=0;
  for(walk=addr_spec; *walk!=0; walk++)
    if (((*walk<'0')||(*walk>'9')) && (*walk!='.')) addr_kind=1;
  if (addr_kind==1) {
    struct hostent *h = gethostbyname(addr_spec);
    if (!h) throw_parse_cb_error("Couldn't resolve address by name", 0);
    memcpy((char*)&addr, h->h_addr, h->h_length);
  } else {
    if (parse_dotted_quad_address(addr_spec, &addr)) return(1);
  }

  /* Try to digest the mask part */
  if (mask_str) {
    addr_kind=0;
    for(walk=mask_str; *walk!=0; walk++) {
      if      (((*walk<'0')||(*walk>'9'))&&(*walk!='.')) addr_kind=2;
      else if ((addr_kind<2)&&(*walk=='.'))              addr_kind=1;
    }
    if (addr_kind==2) {
      struct hostent *h = gethostbyname(mask_str);
      if (!h) throw_parse_cb_error("Couldn't resolve mask address by name",
                                   mask_str - addr_spec);
      memcpy((char*)&mask, h->h_addr, h->h_length);
    } else if (addr_kind==1) {
      if (parse_dotted_quad_address(mask_str, &mask)) {
        parse_cb_error_char += mask_str - addr_spec;
        return(1);
      }
    } else {
      int mask_bits = atoi(mask_str);
      mask.s_addr = htonl(((1<<mask_bits) - 1) << (32-mask_bits));
    }
  } else
    mask.s_addr = INADDR_BROADCAST;

  /* Parse port range */
  if (ports_str) {
    char *range_end_str = 0;
    for(walk=ports_str; *walk!=0; walk++) {
      if (((*walk<'0')||(*walk>'9')) && (*walk!='-'))
        throw_parse_cb_error("Failed to parse port range", -1)
      else if (*walk=='-') {
        *walk = 0;
        if (range_end_str)
          throw_parse_cb_error("Found second '-' in port range", -1);
        range_end_str = walk + 1;
      }
    }
    port_range_start = atoi(ports_str);
    if (range_end_str)
      port_range_end = atoi(range_end_str);
    else
      port_range_end = port_range_start;
  } else {
    port_range_start = 0;
    port_range_end   = 0xffff;
  }

  /* Insert this allow/deny policy into the list */
  new = (struct INET_allow_deny*)malloc(sizeof(struct INET_allow_deny));
  if (!new) die("Failed to allocate memory for allow/deny list!");
  new->AD_policy        = AD_policy;
  new->addr             = addr;
  new->mask             = mask;
  new->port_range_start = port_range_start;
  new->port_range_end   = port_range_end;
  new->next             = 0;
  if (INET_AD_tail==0) { INET_AD_list = INET_AD_tail = new;            }
  else                 { INET_AD_tail->next = new; INET_AD_tail = new; }

  /* Dump the rule to the log file */
  if (debug_rules) {
    char *addr_str = strdup(inet_ntoa(addr));
    char *mask_str = strdup(inet_ntoa(mask));
    if (addr_str && mask_str) {
      log_msg("INET rule: %s to %s/%s:%d-%d\n",
              (AD_policy==ALLOW)?"Allow":"Deny",
              addr_str, mask_str, port_range_start, port_range_end);
      free(addr_str);
      free(mask_str);
    } else
      die("Failed to allocate memory during initialization.");
  }

  return(0);
}

int parse_UNIX_AD_cb(AD_policy_t AD_policy, char *path) {
  struct UNIX_allow_deny *new;
  new = (struct UNIX_allow_deny*)malloc(sizeof(struct UNIX_allow_deny));
  if (!new) die("Failed to allocate memory for allow/deny list!");
  new->AD_policy = AD_policy;
  if (!strncmp(path, "ALL", 4)) {
    new->all = 1;
    new->path = 0;
  } else {
    new->all = 0;
    new->path = strdup(path);
    if (!new->path) die("Failed to allocate memory for allow/deny list!");
  }
  new->next = 0;
  if (UNIX_AD_tail==0) UNIX_AD_list = UNIX_AD_tail = new;
  else                 { UNIX_AD_tail->next = new; UNIX_AD_tail = new; }

  /* Dump the rule to the log file */
  if (debug_rules) {
    log_msg("UNIX rule: %s to %s\n",
            (AD_policy==ALLOW)?"Allow":"Deny", path);
  }
  return(0);
}

void parse_allow_deny_string(char *AD_str, char *AFstr,
                             int(*cb)(AD_policy_t, char*))
{
  char parse_err[128];
  AD_policy_t  AD_policy = -1;
  char       *spec_str = 0;
  char       *walk = AD_str;
  int         escape = 0;
  int         shift = 0;
  int         state = 0;

  sprintf(parse_err, "Failed to parse $NETJAIL_%s allow/deny environment " \
                     "variable.\n", AFstr);

  while (state<3) {
    switch(state) {
      case 0:
        if      (*walk=='A') { AD_policy = ALLOW; state=1; }
        else if (*walk=='D') { AD_policy = DENY;  state=1; }
        else
          die("%sExpected 'A' or 'D' at character %d",
              parse_err, walk - AD_str + 1);
        break;
      case 1:
        if (*walk==':') { state=2; }
        else
          die("%sExpected ':' at character %d", parse_err, walk - AD_str + 1);
        break;
      case 2:
        if (spec_str==0) { spec_str = walk; }
        if      (*walk==0)                  { state=3; }
        else if ((*walk==',') && (!escape)) { state=0; *walk=0; }

        if (*walk=='\\') { escape=1; shift++; }
        else             { escape=0; *(walk-shift) = *walk; }

        if (state!=2) {
          if (strlen(spec_str)==0)
            die("%sAddress specification was empty at character %d",
                parse_err, walk - AD_str - 1);

          if (cb(AD_policy, spec_str))
            die("%s%s at character %d", parse_err, parse_cb_error,
                spec_str - AD_str + parse_cb_error_char + 1);
          AD_policy = -1;
          spec_str = 0;
          escape = shift = 0;
        }
        break;
      default: state=3; break; /* Can't happen, but let's fix it anyway... */
    }
    walk++;
  }
}

void lib_init() {
  void *libhdl;
  char *dlerr;
  char *logfile, default_log[]="netjail.log";
  char *UNIX_AD_env, *INET_AD_env;
  char *debug_rules_env;
  int i;

  if (lib_initialized) return;

  if (!(libhdl=dlopen(SOCKET_LIB, RTLD_LAZY)))
    die("Failed to patch library calls: %s", dlerror());

  orig_socket = dlsym(libhdl, "socket");
  if ((dlerr=dlerror()) != NULL)
    die("Failed to patch socket() library call: %s", dlerr);

  orig_connect = dlsym(libhdl, "connect");
  if ((dlerr=dlerror()) != NULL)
    die("Failed to patch connect() library call: %s", dlerr);

  if (!(logfile=getenv("NETJAIL_LOG"))) {
    logfile = default_log;
  }
  if (!(log = fopen(logfile, "a")))
    die("Failed to open log file (%s): %s", logfile, strerror(errno));

  if ((debug_rules_env=getenv("NETJAIL_DEBUG")))
    debug_rules = atoi(debug_rules_env);

  lib_initialized = -1;

  if ((INET_AD_env=getenv("NETJAIL_INET")))
    parse_allow_deny_string(INET_AD_env, "INET", parse_INET_AD_cb);

  if ((UNIX_AD_env=getenv("NETJAIL_UNIX")))
    parse_allow_deny_string(UNIX_AD_env, "UNIX", parse_UNIX_AD_cb);

  for(i=0; i<32; i++) { INETfds[i] = 0; }
  for(i=0; i<32; i++) { UNIXfds[i] = 0; }

  lib_initialized = 1;
}

static char *proto_family[] = {
  "UNSPEC", "LOCAL", "INET", "AX25", "IPX", "APPLETALK", "NETROM", "BRIDGE",
  "ATMPVC", "X25", "INET6", "ROSE", "DECnet", "NETBEUI", "SECURITY", "KEY",
  "NETLINK", "PACKET", "ASH", "ECONET", "ATMSVC", "UNKNOWN", "SNA", "IRDA",
  "PPPOX", "WANPIPE", "UNKNOWN", "UNKNOWN", "UNKNOWN", "UNKNOWN", "UNKNOWN",
  "BLUETOOTH"
};

static char *socket_type[] = {
  "UNSPEC", "STREAM", "DGRAM", "RAW", "RDM", "SEQPACKET", "UNKNOWN",
  "UNKNOWN", "UNKNOWN", "UNKNOWN", "PACKET"
};

int socket(int family, int type, int proto) {
  int fd;
  struct protoent *p_ent;
  char *p_name;
  char unkwn[] = "Unknown";

  lib_init();

  fd = orig_socket(family, type, proto);
  p_ent = getprotobynumber(proto);
  if (p_ent) { p_name = p_ent->p_name; } else { p_name = unkwn; }
  log_msg("socket(PF_%s, SOCK_%s, %s) = %d\n",
          proto_family[family], socket_type[type], p_name, fd);

  if      (family == PF_INET) { addINETfd(fd); }
  else if (family == PF_UNIX) { addUNIXfd(fd); }

  return(fd);
}

AD_policy_t check_INET_policy(struct in_addr addr, in_port_t port) {
  AD_policy_t policy;
  struct INET_allow_deny *p;
  if (lib_initialized == -1) return(ALLOW);
  policy = ALLOW;
  for(p=INET_AD_list; p; p=p->next)
    if (((addr.s_addr & p->mask.s_addr)==(p->addr.s_addr & p->mask.s_addr)) &&
        ((port >= p->port_range_start) && (port <= p->port_range_end)))
    {
      policy = p->AD_policy;
      if (debug_rules) {
        char *addr_str = strdup(inet_ntoa(p->addr));
        char *mask_str = strdup(inet_ntoa(p->mask));
        if (addr_str && mask_str) {
          log_msg("Match of rule: policy=%s, %s/%s:%d-%d\n",
                  (policy==ALLOW)?"allow":"deny",
                  addr_str, mask_str, p->port_range_start, p->port_range_end);
          free(addr_str);
          free(mask_str);
        }
      }
    }
  return(policy);
}

AD_policy_t check_UNIX_policy(char *path) {
  AD_policy_t policy;
  struct UNIX_allow_deny *p;
  if (lib_initialized == -1) return(ALLOW);
  policy = ALLOW;
  for(p=UNIX_AD_list; p; p=p->next)
    if ((p->all) || (!strcmp(p->path, path))) {
      policy = p->AD_policy;
      if (debug_rules) {
        log_msg("Match of rule: policy=%s, %s\n",
                (policy==ALLOW)?"Allow":"Deny", (p->all)?"ALL":p->path);
      }
    }
  return(policy);
}

int connect(int sockfd, const struct sockaddr* addr, socklen_t addr_len) {
  int retval;
  char *addr_str;
  char errstr[128];
  int prevent_connect = 0;

  lib_init();

  if (isINETfd(sockfd)) {
    struct in_addr inaddr = ((struct sockaddr_in*)addr)->sin_addr;
    in_port_t      inport = ntohs(((struct sockaddr_in*)addr)->sin_port);
    addr_str = malloc(32);
    sprintf(addr_str, "%s:%hu", inet_ntoa(inaddr), inport);
    prevent_connect = (check_INET_policy(inaddr, inport) == DENY);
  } else if (isUNIXfd(sockfd)) {
    addr_str = strdup(((struct sockaddr_un*)addr)->sun_path);
    prevent_connect = (check_UNIX_policy(addr_str) == DENY);
  } else {
    addr_str = strdup("Address not AF_INET or AF_UNIX");
  }

  if (!addr_str)
    die("Couldn't get enough memory for logging!  I'm just " \
        "shutting everything down so that we don't accidentally " \
        "let something through unnoticed...\n");

  if (prevent_connect) {
    retval = -1;
    errno = ECONNREFUSED;
  } else
    retval = orig_connect(sockfd, addr, addr_len);

  if (retval)
    sprintf(errstr, " (%s)", strerror(errno));
  else
    errstr[0] = 0;

  log_msg("connect(%d, %s, %d) = %d%s\n",
          sockfd, addr_str, addr_len, retval, errstr);
  free(addr_str);

  return(retval);
}
